import jax.numpy as jnp
import equinox as eqx

from . import const as const 
from .reactions import Reaction

class NuclearRates(eqx.Module): 
    """
    Nuclear net and rates. 

    Attributes
    ----------
    max_i_species : int
        Number of nuclear species to track. Convention is 0:n, 1:p, 2:d, 3:t,
        4:He3, 5:a, 6:Li7, 7:Be7, 8: Li6, 9: He6, 10: Li8, 11: B8. 
        max_i_species = 8 would go up to Be7, for example. 
    interp_type : str
        Interpolation method for nuclear rates with spline data. Either 'linear'
        or 'log'. 
    reactions : list of Reaction
        Reactions to consider in the network. 
    in_states : dict of tuple
        Nuclear species on the LHS of each reaction. 
    out_states : dict of tuple
        Nuclear species on the RHS of each reaction. 
    frwrd_symmetry_fac : dict of float
        Symmetry factor for forward direction of reaction. 
    bkwrd_symmetry_fac : dict of float
        Symmetry factor for backward direction of reaction. 
    frwrd_rate : dict of callable
        Dictionary of forward rate parameter (Gamma, <sigma v> or 
        <sigma v^2> divided by (1 amu)^(N_in-1)) for each reaction, 
        units (s^-1/g, cm^3/s/g or cm^6/s/g^2)
    bkwrd_rate : dict of callable
        Dictionary of backward rate parameter for each reaction. 
    """
    
    max_i_species : int
    interp_type : str
    reactions : list
    reactions_names: list 
    in_states : dict
    out_states : dict
    frwrd_symmetry_fac : dict
    bkwrd_symmetry_fac : dict 
    frwrd_rate_param : dict 
    bkwrd_rate_param : dict
    frwrd_reaction_by_particle : dict 
    bkwrd_reaction_by_particle : dict

    def __init__(
        self, reactions=None, nuclear_net=None, interp_type='linear',
        max_i_species=None
    ):
        """
        Populates the class. 

        Parameters
        ----------
        reactions : list of Reaction
            Reactions to be considered in the nuclear net. 
        nuclear_net : str
            Used for fixed networks, choices are: 
            'np_only': Only n <-> p rates will be used. 
            'key_PRIMAT_2018': PRIMAT 2018 small network rates, also used 
            by PRyMordial. 
            'key_YOF': YOF small network rates, also used by 
            PRyMordial.
            'key_PRIMAT_2023': PRIMAT 2022 rates.
            'key_PArthENoPE': PArthENoPE rates. 
            'full_PRIMAT_2022': PRIMAT 2022 full network rates (unverified). 
        interp_type : str
            Interpolation type for nuclear rates with spline data. Either 
            'linear' or 'log'. 
        max_i_species : int
            Number of nuclear species to consider. 
        """
 
        self.interp_type = interp_type
        self.max_i_species = max_i_species

        if reactions is not None: 

            self.reactions = tuple(reactions)

        elif nuclear_net == 'np_only': 
            # No nuclear reactions. n<->p rates are always included. 

            self.reactions = ()
            self.max_i_species = 2

        else: 

            self.reactions = tuple(self.populate(nuclear_net))

            if nuclear_net[:3] == 'key': 
 
                self.max_i_species = 8 

            elif nuclear_net[:5] == 'small': 

                self.max_i_species = 9

            else: 

                self.max_i_species = 12
            
        self.in_states = {} 
        self.out_states = {} 
        self.frwrd_symmetry_fac = {} 
        self.bkwrd_symmetry_fac = {} 
        self.frwrd_rate_param = {} 
        self.bkwrd_rate_param = {}
        self.reactions_names = ()

        self.frwrd_reaction_by_particle = {
            i:[] for i in range(self.max_i_species)
        }
        self.bkwrd_reaction_by_particle = {
            i:[] for i in range(self.max_i_species)
        }

        for rxn in self.reactions: 

            self.in_states[rxn.name]  = rxn.in_states
            self.out_states[rxn.name] = rxn.out_states 
            self.frwrd_symmetry_fac[rxn.name] = rxn.frwrd_symmetry_fac 
            self.bkwrd_symmetry_fac[rxn.name] = rxn.bkwrd_symmetry_fac 
            self.frwrd_rate_param[rxn.name] = rxn.frwrd_rate_param
            self.bkwrd_rate_param[rxn.name] = rxn.bkwrd_rate_param

            for i in range(self.max_i_species): 
                if i in self.in_states[rxn.name]:
                    self.frwrd_reaction_by_particle[i].append(rxn.name)
                if i in self.out_states[rxn.name]: 
                    self.bkwrd_reaction_by_particle[i].append(rxn.name)
            
            self.reactions_names = self.reactions_names + (rxn.name,)
            

    @eqx.filter_jit
    def __call__(
        self, Y, T_t, rhoBBN, T_interval, 
        nTOp_frwrd_vec, nTOp_bkwrd_vec, tau_n_fac=1., nuclear_rates_q=None
    ): 
        """
        Returns the rate of change of the abundances.  

        Parameters
        ----------
        Y : Array
            Current abundance of species (n_species/n_b).
        T_t : float
            Current temperature in MeV. 
        rhoBBN : float
            Number density of baryons x 1amu in g/cm^3. 
        T_interval : array of float
            Interval over which n<->p rates were calculated in MeV. 
        nTOp_frwrd_vec : array of float
            n -> p dimensionless rates corresponding to T_interval, normalized 
            to the neutron decay width, i.e. Yn * nTOp_frwrd / tau_n + ... 
            = -dYn/dt. 
        nTOp_bkwrd_vec : array of float
            p -> n dimensionless rates corresponding to T_interval. 
        tau_n_fac : float
            Rescaling parameter for neutron decay lifetime. 
        nuclear_rates_q : array
            Rescaling parameter of expsigma in nuclear rate. If None, 
            no rescaling is assumed.

        Returns
        -------
        Array
            dY/dt in s^-1. Same dimensions as Y. 
        """

        if nuclear_rates_q is None: 

            nuclear_rates_q = jnp.array([0. for _ in self.reactions])

        dYdt_vec = jnp.zeros(len(Y))
        
        _nTOp_frwrd = jnp.interp(
            T_t, jnp.flip(T_interval), jnp.flip(nTOp_frwrd_vec),
            left=nTOp_frwrd_vec[-1],right=nTOp_frwrd_vec[0]
        ) / (const.tau_n * tau_n_fac)

        _nTOp_bkwrd = jnp.interp(
            T_t, jnp.flip(T_interval), jnp.flip(nTOp_bkwrd_vec),
            left=nTOp_bkwrd_vec[-1],right=nTOp_bkwrd_vec[0]
        ) / (const.tau_n * tau_n_fac)

        # These functions take temperature in K. 
        frwrd_rate_params = {
            rxn.name:self.frwrd_rate_param[rxn.name](
                T_t / const.kB, nuclear_rates_q[i]
            ) for i,rxn in enumerate(self.reactions)
        } 
        bkwrd_rate_params = {
            rxn.name:self.bkwrd_rate_param[rxn.name](
                T_t / const.kB, nuclear_rates_q[i]
            ) for i,rxn in enumerate(self.reactions)
        }

        dYdt_vec = dYdt_vec.at[0].set(-_nTOp_frwrd*Y[0] + _nTOp_bkwrd*Y[1])
        dYdt_vec = dYdt_vec.at[1].set( _nTOp_frwrd*Y[0] - _nTOp_bkwrd*Y[1])

        for rxn in self.reactions: 
            dYdt_vec += self.get_dYdt_rxn(
                rxn.name, Y, rhoBBN, 
                frwrd_rate_params[rxn.name], bkwrd_rate_params[rxn.name]
            )

        return tuple(dYdt_vec)
    
    def get_dYdt_rxn(
        self, rxn, Y, rhoBBN, frwrd_rate_param, bkwrd_rate_param
    ):
        """
        Returns the rate of change of abundances due to a particular reaction.

        Parameters
        ----------
        rxn : str
            Name of the reaction. 
        Y : Array
            Current abundance of species (n_species/n_b).
        rhoBBN : float
            Number density of baryons x 1amu in g/cm^3. 
        frwrd_rate_param : callable
            Function returning the forward rate (either <sigma v> or 
            <sigma v^2>). Takes two arguments, `T` for EM temperature *in K* and `p` for rescaling of the rate. 
        bkwrd_rate_param : callable
            Similar to `frwrd_rate_param`, but for the backward rate. 

        Returns
        -------
        Array
            dY/dt in s^-1. Same dimensions as Y. 
        """
        
        dYdt_vec = jnp.zeros(len(Y))
        Y_prod_frwrd = 1. 
        Y_prod_bkwrd = 1. 
        for i in self.in_states[rxn]: 
            Y_prod_frwrd *= Y[i]
        for j in self.out_states[rxn]: 
            Y_prod_bkwrd *= Y[j] 

        for i in self.in_states[rxn]: 
            dYdt_vec = dYdt_vec.at[i].add(
                - self.frwrd_symmetry_fac[rxn]
                * rhoBBN**(len(self.in_states[rxn])-1)
                * Y_prod_frwrd * frwrd_rate_param 
            )
            dYdt_vec = dYdt_vec.at[i].add(
                + self.bkwrd_symmetry_fac[rxn] 
                * rhoBBN**(len(self.out_states[rxn])-1)
                * Y_prod_bkwrd * bkwrd_rate_param
            )

        for j in self.out_states[rxn]: 
            dYdt_vec = dYdt_vec.at[j].add(
                + self.frwrd_symmetry_fac[rxn]
                * rhoBBN**(len(self.in_states[rxn])-1)
                * Y_prod_frwrd * frwrd_rate_param 
            )
            dYdt_vec = dYdt_vec.at[j].add(
                - self.bkwrd_symmetry_fac[rxn] 
                * rhoBBN**(len(self.out_states[rxn])-1)
                * Y_prod_bkwrd * bkwrd_rate_param
            )

        return dYdt_vec
    
    def populate(self, network): 
        """
        Populate the nuclear rates. 

        Parameters
        ----------
        network : str
            Nuclear network of interest. Choices are: 
            'key_PRIMAT_2018': PRIMAT 2018 key network rates, also used 
            by PRyMordial. 
            'key_YOF': YOF key network rates, also used by 
            PRyMordial.
            'key_PRIMAT_2023': PRIMAT 2023 key network rates.
            'key_PArthENoPE': PArthENoPE key network rates. 
            'full_PRIMAT_2023': PRIMAT 2023 full network rates.

        Returns
        -------
        list of Reaction
        """

        key_rxns_dict = {} 

        if network != 'key_PArthENoPE': 

            if (
                network == 'key_PRIMAT_2018' 
                or network == 'key_YOF' 
                or network == 'key_PRIMAT_2023'
            ): 
                key_network_str = network 
            elif network == 'full_PRIMAT_2018': 
                key_network_str = 'key_PRIMAT_2018'
            elif (
                network == 'small_PRIMAT_2023'
                or network == 'full_PRIMAT_2023'
            ): 
                key_network_str = 'key_PRIMAT_2023'

            key_rxns_dict['npdg'] = Reaction(
                'npdg', (0, 1), (2, ), 4.7161402e9, 1.5, -25.81502, 
                spline_data=key_network_str+'/npdg.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['dpHe3g'] = Reaction(
                'dpHe3g', (1, 2), (4, ), 1.6335102e10, 1.5, -63.749132, 
                spline_data=key_network_str+'/dpHe3g.txt',
                interp_type=self.interp_type
            )
            key_rxns_dict['ddHe3n'] = Reaction(
                'ddHe3n', (2, 2), (0, 4), 1.7318296, 0., -37.934112, 
                spline_data=key_network_str+'/ddHe3n.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['ddtp'] = Reaction(
                'ddtp', (2, 2), (1, 3), 1.7349209, 0., -46.797116, 
                spline_data=key_network_str+'/ddtp.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['tpag'] = Reaction(
                'tpag', (1, 3), (5, ), 2.610575e10, 1.5, -229.93039, 
                spline_data=key_network_str+'/tpag.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['tdan'] = Reaction(
                'tdan', (2, 3), (0, 5), 5.5354059, 0., -204.11537, 
                spline_data=key_network_str+'/tdan.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['taLi7g'] = Reaction(
                'taLi7g', (3, 5), (6, ), 1.1132988e10, 1.5, -28.635551, 
                spline_data=key_network_str+'/taLi7g.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['He3ntp'] = Reaction(
                'He3ntp', (0, 4), (1, 3), 1.001785, 0., -8.8630042, 
                spline_data=key_network_str+'/He3ntp.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['He3dap'] = Reaction(
                'He3dap', (2, 4), (1, 5), 5.5452865, 0., -212.97837, 
                spline_data=key_network_str+'/He3dap.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['He3aBe7g'] = Reaction(
                'He3aBe7g', (4, 5), (7, ), 1.1128943e10, 1.5, -18.417922, 
                spline_data=key_network_str+'/He3aBe7g.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['Be7nLi7p'] = Reaction(
                'Be7nLi7p', (0, 7), (1, 6), 1.0021491, 0., -19.080633, 
                spline_data=key_network_str+'/Be7nLi7p.txt', 
                interp_type=self.interp_type
            )
            key_rxns_dict['Li7paa'] = Reaction(
                'Li7paa', (1, 6), (5, 5), 4.6898011, 0., -201.29484, 
                spline_data=key_network_str+'/Li7paa.txt', 
                interp_type=self.interp_type
            )
            

        else: 

            from .data.nuclear_rates import key_PArthENoPE as PArth

            key_rxns_dict['npdg'] = Reaction(
                'npdg', (0, 1), (2, ), 4.7161402e9, 1.5, -25.81502,
                frwrd_rate_param_func = PArth.npdg_frwrd_rate
            )
            key_rxns_dict['dpHe3g'] = Reaction(
                'dpHe3g', (1, 2), (4, ), 1.6335102e10, 1.5, -63.749132,  
                frwrd_rate_param_func = PArth.dpHe3g_frwrd_rate
            )
            key_rxns_dict['ddHe3n'] = Reaction(
                'ddHe3n', (2, 2), (0, 4), 1.7318296, 0., -37.934112, 
                frwrd_rate_param_func = PArth.ddHe3n_frwrd_rate
            )
            key_rxns_dict['ddtp'] = Reaction(
                'ddtp', (2, 2), (1, 3), 1.7349209, 0., -46.797116, 
                frwrd_rate_param_func = PArth.ddtp_frwrd_rate
            )
            key_rxns_dict['tpag'] = Reaction(
                'tpag', (1, 3), (5, ), 2.610575e10, 1.5, -229.93039,  
                frwrd_rate_param_func = PArth.tpag_frwrd_rate
            )
            key_rxns_dict['tdan'] = Reaction(
                'tdan', (2, 3), (0, 5), 5.5354059, 0., -204.11537, 
                frwrd_rate_param_func = PArth.tdan_frwrd_rate
            )
            key_rxns_dict['taLi7g'] = Reaction(
                'taLi7g', (3, 5), (6, ), 1.1132988e10, 1.5, -28.635551, 
                frwrd_rate_param_func = PArth.taLi7g_frwrd_rate
            )
            key_rxns_dict['He3ntp'] = Reaction(
                'He3ntp', (0, 4), (1, 3), 1.001785, 0., -8.8630042, 
                frwrd_rate_param_func = PArth.He3ntp_frwrd_rate
            )
            key_rxns_dict['He3dap'] = Reaction(
                'He3dap', (2, 4), (1, 5), 5.5452865, 0., -212.97837, 
                frwrd_rate_param_func = PArth.He3dap_frwrd_rate
            )
            key_rxns_dict['He3aBe7g'] = Reaction(
                'He3aBe7g', (4, 5), (7, ), 1.1128943e10, 1.5, -18.417922, 
                frwrd_rate_param_func = PArth.He3aBe7g_frwrd_rate
            )
            key_rxns_dict['Be7nLi7p'] = Reaction(
                'Be7nLi7p', (0, 7), (1, 6), 1.0021491, 0., -19.080633, 
                frwrd_rate_param_func = PArth.Be7nLi7p_frwrd_rate
            )
            key_rxns_dict['Li7paa'] = Reaction(
                'Li7paa', (1, 6), (5, 5), 4.6898011, 0., -201.29484, 
                frwrd_rate_param_func = PArth.Li7paa_frwrd_rate
            )


        if network[:4] == 'full': 

            other_rxns_dict = {}

            dir_name = 'other_PRIMAT_2023/'

            other_rxns_dict['Li7paag'] = Reaction(
                'Li7paag', (1, 6), (5, 5), 4.6898, 0.,-201.295, 
                spline_data=dir_name+'Li7paag.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7naa'] = Reaction(
                'Be7naa', (0, 7), (5, 5), 4.6982, 0.,-220.3871, 
                spline_data=dir_name+'Be7naa.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7daap'] = Reaction(
                'Be7daap', (2, 7), (1, 5, 5), 9.9579e-10, -1.5, -194.5722, 
                spline_data=dir_name+'Be7daap.txt',
                interp_type=self.interp_type
            ) 
            other_rxns_dict['daLi6g'] = Reaction(
                'daLi6g', (2, 5), (8, ), 1.53053e10, 1.5, -17.1023,
                spline_data=dir_name+'daLi6g.txt',
                interp_type=self.interp_type
            ) 
            other_rxns_dict['Li6pBe7g'] = Reaction(
                'Li6pBe7g', (1, 8), (7, ), 1.18778e10, 1.5, -65.0648, 
                spline_data=dir_name+'Li6pBe7g.txt',
                interp_type=self.interp_type
            ) 
            other_rxns_dict['Li6pHe3a'] = Reaction(
                'Li6pHe3a', (1, 8), (4, 5), 1.06729, 0., -46.6469, 
                spline_data=dir_name+'Li6pHe3a.txt',
                interp_type=self.interp_type
            ) 
            other_rxns_dict['B8naap'] = Reaction(
                'B8naap', (0, 11), (1, 5, 5), 3.6007e-10, -1.5, -218.7915, 
                spline_data=dir_name+'B8naap.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6He3aap'] = Reaction(
                'Li6He3aap', (4, 8), (1, 5, 5), 7.2413e-10, -1.5, -195.8748, 
                spline_data=dir_name+'Li6He3aap.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6taan'] = Reaction(
                'Li6taan', (3, 8), (0, 5, 5), 7.2333e-10, -1.5, -187.0131, 
                spline_data=dir_name+'Li6taan.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6tLi8p'] = Reaction(
                'Li6tLi8p', (3, 8), (1, 10), 2.0167, 0., -9.306, 
                spline_data=dir_name+'Li6tLi8p.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li7He3Li6a'] = Reaction(
                'Li7He3Li6a', (4, 6), (5, 8), 2.1972, 0., -154.6607, 
                spline_data=dir_name+'Li7He3Li6a.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li8He3Li7a'] = Reaction(
                'Li8He3Li7a', (4, 10), (5, 6), 1.9994, 0., -215.2055, 
                spline_data=dir_name+'Li8He3Li7a.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7tLi6a'] = Reaction(
                'Be7tLi6a', (3, 7), (5, 8), 2.1977, 0., -164.8783, 
                spline_data=dir_name+'Be7tLi6a.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['B8tBe7a'] = Reaction(
                'B8tBe7a', (3, 11), (5, 7), 1.9999, 0., -228.3344, 
                spline_data=dir_name+'B8tBe7a.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['B8nLi6He3'] = Reaction(
                'B8nLi6He3', (0, 11), (4, 8), 0.49669, 0., -22.9167, 
                spline_data=dir_name+'B8nLi6He3.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['B8nBe7d'] = Reaction(
                'B8nBe7d', (0, 11), (2, 7), 0.36119, 0., -24.2194, 
                spline_data=dir_name+'B8nBe7d.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6tLi7d'] = Reaction(
                'Li6tLi7d', (3, 8), (2, 6), 0.72734, 0., -11.5332, 
                spline_data=dir_name+'Li6tLi7d.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6He3Be7d'] = Reaction(
                'Li6He3Be7d', (4, 8), (2, 7), 0.72719, 0., -1.3157, 
                spline_data=dir_name+'Li6He3Be7d.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li7He3aad'] = Reaction(
                'Li7He3aad', (4, 6), (2, 5, 5), 2.8700e-10, -1.5, -137.5575, 
                spline_data=dir_name+'Li7He3aad.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li8He3aat'] = Reaction(
                'Li8He3aat', (4, 10), (3, 5, 5), 3.5907e-10, -1.5, -186.5821, 
                spline_data=dir_name+'Li8He3aat.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7taad'] = Reaction(
                'Be7taad', (3, 7), (2, 5, 5), 2.8706e-10, -1.5, -147.7751, 
                spline_data=dir_name+'Be7taad.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7tLi7He3'] = Reaction(
                'Be7tLi7He3', (3, 7), (4, 6), 1.0002, 0., -10.2176, 
                spline_data=dir_name+'Be7tLi7He3.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['B8dBe7He3'] = Reaction(
                'B8dBe7He3', (2, 11), (4, 7), 1.2514, 0, -62.1535, 
                spline_data=dir_name+'B8dBe7He3.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['B8taaHe3'] = Reaction(
                'B8taaHe3', (3, 11), (4, 5, 5), 3.5922e-10, -1.5, -209.9285, 
                spline_data=dir_name+'B8taaHe3.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7He3ppaa'] = Reaction(
                'Be7He3ppaa', (4, 7), (1, 1, 5, 5), 1.2201e-19, -3., -130.8113, 
                spline_data=dir_name+'Be7He3ppaa.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['ddag'] = Reaction(
                'ddag', (2, 2), (5, ), 4.5310e10, 1.5, -276.7271, 
                spline_data=dir_name+'ddag.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['He3He3app'] = Reaction(
                'He3He3app', (4, 4), (1, 1, 5), 3.3915e-10, -1.5, -149.2290, 
                spline_data=dir_name+'He3He3app.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7pB8g'] = Reaction(
                'Be7pB8g', (1, 7), (11, ), 1.3063e10, 1.5, -1.5825, 
                spline_data=dir_name+'Be7pB8g.txt',
                interp_type=self.interp_type
            ) 
            other_rxns_dict['Li7daan'] = Reaction(
                'Li7daan', (2, 6), (0, 5, 5), 9.9435e-10, -1.5, -175.4916, 
                spline_data=dir_name+'Li7daan.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['dntg'] =   Reaction(
                'dntg', (0, 2), (3, ), 1.6364262e10, 1.5, -72.612132, 
                spline_data=dir_name+'dntg.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['ttann'] =   Reaction(
                'ttann', (3, 3), (0, 0, 5), 3.3826187e-10, -1.5, -131.50322, 
                spline_data=dir_name+'ttann.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['He3nag'] = Reaction(
                'He3nag', (0, 4), (5, ), 2.6152351e10, 1.5, -238.79338, 
                spline_data=dir_name+'He3nag.txt',
                interp_type=self.interp_type    
            )
            other_rxns_dict['He3tad'] = Reaction(
                'He3tad', (3, 4), (2, 5), 1.5981381, 0., -166.18124, 
                spline_data=dir_name+'He3tad.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['He3tanp'] = Reaction(
                'He3tanp', (3, 4), (0, 1, 5), 3.3886566e-10, -1.5, -140.36623, 
                spline_data=dir_name+'He3tanp.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li7taann'] = Reaction(
                'Li7taann', (3, 6), (0, 0, 5, 5), 
                1.2153497e-19, -3., -102.86767, 
                spline_data=dir_name+'Li7taann.txt',
                interp_type=self.interp_type
            )                                               
            other_rxns_dict['Li7He3aanp'] = Reaction(
                'Li7He3aanp', (4, 6), (0, 1, 5, 5), 6.0875952e-20, -3., -111.73068, 
                spline_data=dir_name+'Li7He3aanp.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li8dLi7t'] = Reaction(
                'Li8dLi7t', (2, 10), (3, 6), 1.2509926, 0., -49.02453, 
                spline_data=dir_name+'Li8dLi7t.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7taanp'] = Reaction(
                'Be7taanp', (3, 7), (0, 1, 5, 5), 
                6.0898077e-20, -3., -121.9483, 
                spline_data=dir_name+'Be7taanp.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6nta'] = Reaction(
                'Li6nta', (0, 8), (3, 5), 1.0691921, 0., -55.509875, 
                spline_data=dir_name+'Li6nta.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['He3tLi6g'] = Reaction(
                'He3tLi6g', (3, 4), (8, ), 2.4459918e10, 1.5, -183.2835, 
                spline_data=dir_name+'He3tLi6g.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['anpLi6g'] = Reaction(
                'anpLi6g', (0, 1, 5), (8, ), 7.2181753e19, 3., -42.917276, 
                spline_data=dir_name+'anpLi6g.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6nLi7g'] = Reaction(
                'Li6nLi7g', (0, 8), (6, ), 1.1903305e10, 1.5, -84.145424, 
                spline_data=dir_name+'Li6nLi7g.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6dLi7p'] = Reaction(
                'Li6dLi7p', (2, 8), (1, 6), 2.5239503, 0., -58.330405, 
                spline_data=dir_name+'Li6dLi7p.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6dBe7n'] = Reaction(
                'Li6dBe7n', (2, 8), (0, 7), 2.5185377, 0., -39.249773, 
                spline_data=dir_name+'Li6dBe7n.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li7nLi8g'] = Reaction(
                'Li7nLi8g', (0, 6), (10, ), 1.3081022e10, 1.5, -23.587602, 
                spline_data=dir_name+'Li7nLi8g.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li7dLi8p'] = Reaction(
                'Li7dLi8p', (2, 6), (1, 10), 2.7736709, 0., 2.2274166, 
                spline_data=dir_name+'Li7dLi8p.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li8paan'] = Reaction(
                'Li8paan', (1, 10), (0, 5, 5), 3.5851946e-10, -1.5, -177.70722, 
                spline_data=dir_name+'Li8paan.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['annHe6g'] = Reaction(
                'annHe6g', (0, 0, 5), (9, ), 1.0837999e20, 3., -11.319626, 
                spline_data=dir_name+'annHe6g.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['ppndp'] = Reaction(
                'ppndp', (1, 1, 0), (1, 2), 2.3580703e9, 1.5, -25.815019, 
                spline_data=dir_name+'ppndp.txt',
                interp_type=self.interp_type
            )

        if network[:5] == 'small': 

            other_rxns_dict = {}

            dir_name = 'other_PRIMAT_2023/' 

            other_rxns_dict['Li7paag'] = Reaction(
                'Li7paag', (1, 6), (5, 5), 4.6898, 0.,-201.295, 
                spline_data=dir_name+'Li7paag.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7naa'] = Reaction(
                'Be7naa', (0, 7), (5, 5), 4.6982, 0.,-220.3871, 
                spline_data=dir_name+'Be7naa.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7daap'] = Reaction(
                'Be7daap', (2, 7), (1, 5, 5), 9.9579e-10, -1.5, -194.5722, 
                spline_data=dir_name+'Be7daap.txt',
                interp_type=self.interp_type
            ) 
            other_rxns_dict['daLi6g'] = Reaction(
                'daLi6g', (2, 5), (8, ), 1.53053e10, 1.5, -17.1023,
                spline_data=dir_name+'daLi6g.txt',
                interp_type=self.interp_type
            ) 
            other_rxns_dict['Li6pBe7g'] = Reaction(
                'Li6pBe7g', (1, 8), (7, ), 1.18778e10, 1.5, -65.0648, 
                spline_data=dir_name+'Li6pBe7g.txt',
                interp_type=self.interp_type
            ) 
            other_rxns_dict['Li6pHe3a'] = Reaction(
                'Li6pHe3a', (1, 8), (4, 5), 1.06729, 0., -46.6469, 
                spline_data=dir_name+'Li6pHe3a.txt',
                interp_type=self.interp_type
            ) 
            other_rxns_dict['Li6He3aap'] = Reaction(
                'Li6He3aap', (4, 8), (1, 5, 5), 7.2413e-10, -1.5, -195.8748, 
                spline_data=dir_name+'Li6He3aap.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6taan'] = Reaction(
                'Li6taan', (3, 8), (0, 5, 5), 7.2333e-10, -1.5, -187.0131, 
                spline_data=dir_name+'Li6taan.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li7He3Li6a'] = Reaction(
                'Li7He3Li6a', (4, 6), (5, 8), 2.1972, 0., -154.6607, 
                spline_data=dir_name+'Li7He3Li6a.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7tLi6a'] = Reaction(
                'Be7tLi6a', (3, 7), (5, 8), 2.1977, 0., -164.8783, 
                spline_data=dir_name+'Be7tLi6a.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6tLi7d'] = Reaction(
                'Li6tLi7d', (3, 8), (2, 6), 0.72734, 0., -11.5332, 
                spline_data=dir_name+'Li6tLi7d.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6He3Be7d'] = Reaction(
                'Li6He3Be7d', (4, 8), (2, 7), 0.72719, 0., -1.3157, 
                spline_data=dir_name+'Li6He3Be7d.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li7He3aad'] = Reaction(
                'Li7He3aad', (4, 6), (2, 5, 5), 2.8700e-10, -1.5, -137.5575, 
                spline_data=dir_name+'Li7He3aad.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7taad'] = Reaction(
                'Be7taad', (3, 7), (2, 5, 5), 2.8706e-10, -1.5, -147.7751, 
                spline_data=dir_name+'Be7taad.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7tLi7He3'] = Reaction(
                'Be7tLi7He3', (3, 7), (4, 6), 1.0002, 0., -10.2176, 
                spline_data=dir_name+'Be7tLi7He3.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7He3ppaa'] = Reaction(
                'Be7He3ppaa', (4, 7), (1, 1, 5, 5), 1.2201e-19, -3., -130.8113, 
                spline_data=dir_name+'Be7He3ppaa.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['ddag'] = Reaction(
                'ddag', (2, 2), (5, ), 4.5310e10, 1.5, -276.7271, 
                spline_data=dir_name+'ddag.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['He3He3app'] = Reaction(
                'He3He3app', (4, 4), (1, 1, 5), 3.3915e-10, -1.5, -149.2290, 
                spline_data=dir_name+'He3He3app.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li7daan'] = Reaction(
                'Li7daan', (2, 6), (0, 5, 5), 9.9435e-10, -1.5, -175.4916, 
                spline_data=dir_name+'Li7daan.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['dntg'] =   Reaction(
                'dntg', (0, 2), (3, ), 1.6364262e10, 1.5, -72.612132, 
                spline_data=dir_name+'dntg.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['ttann'] =   Reaction(
                'ttann', (3, 3), (0, 0, 5), 3.3826187e-10, -1.5, -131.50322, 
                spline_data=dir_name+'ttann.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['He3nag'] = Reaction(
                'He3nag', (0, 4), (5, ), 2.6152351e10, 1.5, -238.79338, 
                spline_data=dir_name+'He3nag.txt',
                interp_type=self.interp_type    
            )
            other_rxns_dict['He3tad'] = Reaction(
                'He3tad', (3, 4), (2, 5), 1.5981381, 0., -166.18124, 
                spline_data=dir_name+'He3tad.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['He3tanp'] = Reaction(
                'He3tanp', (3, 4), (0, 1, 5), 3.3886566e-10, -1.5, -140.36623, 
                spline_data=dir_name+'He3tanp.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li7taann'] = Reaction(
                'Li7taann', (3, 6), (0, 0, 5, 5), 
                1.2153497e-19, -3., -102.86767, 
                spline_data=dir_name+'Li7taann.txt',
                interp_type=self.interp_type
            )                                               
            other_rxns_dict['Li7He3aanp'] = Reaction(
                'Li7He3aanp', (4, 6), (0, 1, 5, 5), 6.0875952e-20, -3., -111.73068, 
                spline_data=dir_name+'Li7He3aanp.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Be7taanp'] = Reaction(
                'Be7taanp', (3, 7), (0, 1, 5, 5), 
                6.0898077e-20, -3., -121.9483, 
                spline_data=dir_name+'Be7taanp.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6nta'] = Reaction(
                'Li6nta', (0, 8), (3, 5), 1.0691921, 0., -55.509875, 
                spline_data=dir_name+'Li6nta.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['He3tLi6g'] = Reaction(
                'He3tLi6g', (3, 4), (8, ), 2.4459918e10, 1.5, -183.2835, 
                spline_data=dir_name+'He3tLi6g.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['anpLi6g'] = Reaction(
                'anpLi6g', (0, 1, 5), (8, ), 7.2181753e19, 3., -42.917276, 
                spline_data=dir_name+'anpLi6g.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6nLi7g'] = Reaction(
                'Li6nLi7g', (0, 8), (6, ), 1.1903305e10, 1.5, -84.145424, 
                spline_data=dir_name+'Li6nLi7g.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6dLi7p'] = Reaction(
                'Li6dLi7p', (2, 8), (1, 6), 2.5239503, 0., -58.330405, 
                spline_data=dir_name+'Li6dLi7p.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['Li6dBe7n'] = Reaction(
                'Li6dBe7n', (2, 8), (0, 7), 2.5185377, 0., -39.249773, 
                spline_data=dir_name+'Li6dBe7n.txt',
                interp_type=self.interp_type
            )
            other_rxns_dict['ppndp'] = Reaction(
                'ppndp', (1, 1, 0), (1, 2), 2.3580703e9, 1.5, -25.815019, 
                spline_data=dir_name+'ppndp.txt',
                interp_type=self.interp_type
            )

        if network[:3] == 'key': 

            return list(key_rxns_dict.values())
        
        else: 

            return (
                list(key_rxns_dict.values()) + list(other_rxns_dict.values())
            )
        





import os 

import numpy as np

import jax.numpy as jnp 
import jax.lax as lax
from jax import grad, vmap, device_put, devices

from . import const 
from .special_funcs import Li, K1, K2

###########################################
#                 Cosmology               #
############################################


def Hubble(rho_tot):
    """
    The Hubble parameter in s^-1. 

    Parameters
    ----------
    rho_tot : float or array
        The total energy density in MeV^4. 

    Returns
    -------
    float or array
    """
    
    return (
        rho_tot * 8 * jnp.pi / (3 * const.Mpl**2)
    )**0.5 / const.hbar

def N_eff(rho_tot, rho_g):
    """ Neff parameter. 

    Parameters
    ----------
    rho_tot : float or array
        Total energy density of all fluids. 
    rho_g : float or array
        Energy density of photons. 

    Returns
    -------
    float or array

    """

    return 8./7. * (11./4.)**(4./3.) * (rho_tot-rho_g) / rho_g

def nB(a, eta_fac=1.):
    """ Number density of baryons in MeV^3. 

    Parameters
    ----------
    a : float or array
        Scale factor of interest. 
    eta_fac : float, optional
        Factor to rescale the central value of the baryon-to-photon ratio by.

    Returns
    -------
    float or array

    """
    
    n0B = eta_fac*const.n0CMB*const.eta0 # baryon density of today MeV^3
    return n0B/a**3 # MeV^3


####################################################################
#                 Generic Thermodynamic Variables                  #
####################################################################


def rho_massless_BE(T, mu, g):
    """
    Energy density of a massless particle with Bose-Einstein statistcs.

    Parameters
    ----------
    T : float
        Temperature of massless particle species in units of MeV.
    mu : float
        Chemical potential of massless particle species.
    g : float
        Degrees of freedom of massless particle species (1 for scalar, 
        2 for vector, etc.).

    Returns
    -------
    float
        Units of MeV^4. 
    """
    
    return lax.cond(
        T > 0., 
        lambda T: g * 3 / jnp.pi**2 * Li(4, jnp.exp(mu/T)) * T**4, 
        lambda T: 0.,
        T
    )


def n_massless_BE(T, mu, g): 
    """
    Number density of a massless particle with Bose-Einstein statistcs.

    Parameters
    ----------
    T : float 
        Temperature of massless particle species in units of MeV.
    mu : float
        Chemical potential of massless particle species.
    g : float
        Degrees of freedom of massless particle species (1 for scalar, 
        2 for vector, etc.).

    Returns
    -------
    float
        Units of MeV^3. 
    """

    return lax.cond(
        T > 0., 
        lambda T: g / jnp.pi**2 * Li(3, jnp.exp(mu/T)) * T**3, 
        lambda T: 0., 
        T
    )

def p_massless_BE(T, mu, g): 
    """
    Pressure of a massless particle with Bose-Einstein statistcs.

    Parameters
    ----------
    T : float
        Temperature of massless particle species in units of MeV.
    mu : float
        Chemical potential of massless particle species.
    g : float
        Degrees of freedom of massless particle species (1 for scalar, 
        2 for vector, etc.).

    Returns
    -------
    float
        Units of MeV^4. 
    """

    return rho_massless_BE(T, mu, g) / 3. 

def rho_massless_FD(T, mu, g): 
    """
    Energy density of a massless particle with Fermi-Dirac statistcs.

    Parameters
    ----------
    T : float
        Temperature of massless particle species in units of MeV.
    mu : float
        Chemical potential of massless particle species.
    g : float
        Degrees of freedom of massless particle species (2 for Weyl fermion).

    Returns
    -------
    float
        Units of MeV^4. 
    """

    return lax.cond(
        T > 0., 
        lambda T: -g * 3 / jnp.pi**2 * Li(4, -jnp.exp(mu/T)) * T**4, 
        lambda T: 0., 
        T
    )
 
def n_massless_FD(T, mu, g): 
    """
    Number density of a massless particle with Fermi-Dirac statistcs.

    Parameters
    ----------
    T : float
        Temperature of massless particle species in units of MeV.
    mu : float
        Chemical potential of massless particle species.
    g : float
        Degrees of freedom of massless particle species (2 for Weyl fermion).

    Returns
    -------
    float
        Units of MeV^3. 
    """

    return lax.cond(
        T > 0., 
        lambda T: -g / jnp.pi**2 * Li(3, -jnp.exp(mu/T)) * T**3, 
        lambda T: 0., 
        T
    )

def p_massless_FD(T, mu, g): 
    """
    Pressure of a massless particle with Fermi-Dirac statistcs.

    Parameters
    ----------
    T : float
        Temperature of massless particle species in units of MeV.
    mu : float
        Chemical potential of massless particle species.
    g : float
        Degrees of freedom of massless particle species (2 for Weyl fermion).

    Returns
    -------
    float
        Units of MeV^4. 
    """

    return rho_massless_FD(T, mu, g) / 3. 

def rho_massless_MB(T, mu, g): 
    """
    Energy density of a massless particle with Maxwell-Boltzmann statistcs.

    Parameters
    ----------
    T : float
        Temperature of massless particle species in units of MeV.
    mu : float
        Chemical potential of massless particle species.
    g : float
        Degrees of freedom of massless particle species.

    Returns
    -------
    float
        Units of MeV^4. 
    """

    return lax.cond(
        T > 0., 
        lambda T: g * 3 / jnp.pi**2 * jnp.exp(mu/T) * T**4, 
        lambda T: 0., 
        T
    )

def n_massless_MB(T, mu, g): 
    """
    Number density of a massless particle with Maxwell-Boltzmann statistcs.

    Parameters
    ----------
    T : float
        Temperature of massless particle species in units of MeV.
    mu : float
        Chemical potential of massless particle species.
    g : float
        Degrees of freedom of massless particle species.

    Returns
    -------
    float
        Units of MeV^3. 
    """

    return lax.cond(
        T > 0., 
        lambda T: g / jnp.pi**2 * jnp.exp(mu/T) * T**3, 
        lambda T: 0., 
        T
    )

def p_massless_MB(T, mu, g): 
    """
    Pressure of a massless particle with Maxwell-Boltzmann statistcs.

    Parameters
    ----------
    T : float
        Temperature of massless particle species in units of MeV.
    mu : float
        Chemical potential of massless particle species.
    g : float
        Degrees of freedom of massless particle species.

    Returns
    -------
    float
        Units of MeV^4. 
    """

    return rho_massless_MB(T, mu, g) / 3. 

# Parameters for series approximation of (massive) thermodynamic integrals.
# Series method is much faster than integral computation.
N_series_terms=20 # 20 Seems fine for mu < 0.7 T, << 1% error. 
rel_thres=30.

def rho_massive_BE(T, mu, m, g): 
    """
    Series approximation for energy density of a massive
    particle with Bose-Einstein statistcs.

    Parameters
    ----------
    T : float
        Temperature of massive particle species in MeV.
    mu : float
        Chemical potential of massive particle species in MeV. 
    m : float
        Mass of particle in MeV.
    g : float
        Degrees of freedom of massive particle species (1 for scalar, 
        3 for vector, etc.).

    Returns
    -------
    float
        Units of MeV^4. 
    """
    def res(i, val): 

        return val + jnp.exp(i * mu / T) * (
            (m / T)**3 / i * K1(i * m / T) 
            + 3 * (m / T)**2 / i**2 * K2(i * m / T)
        )

    return jnp.where(
        (T / m > rel_thres) | (T <= 0.), 
        rho_massless_BE(T, mu, g), 
        g / (2 * jnp.pi**2) * T**4 * lax.fori_loop(1, N_series_terms, res, 0)
    )

def n_massive_BE(T, mu, m, g): 
    """
    Series approximation for number density of a massive
    particle with Bose-Einstein statistcs.

    Parameters
    ----------
    T : float
        Temperature of massive particle species in MeV.
    mu : float
        Chemical potential of massive particle species in MeV. 
    m : float
        Mass of particle in MeV.
    g : float
        Degrees of freedom of massive particle species (1 for scalar, 
        3 for vector, etc.).

    Returns
    -------
    float
        Units of MeV^3. 
    """

    def res(i, val): 

        return val + jnp.exp(i * mu / T) * (m / T)**2 / i * K2(i * m / T)

    return jnp.where(
        (T / m > rel_thres) | (T <= 0.), 
        n_massless_BE(T, mu, g), 
        g / (2 * jnp.pi**2) * T**3 * lax.fori_loop(1, N_series_terms, res, 0)
    )


def p_massive_BE(T, mu, m, g): 
    """
    Series approximation for pressure of a massive
    particle with Bose-Einstein statistcs.

    Parameters
    ----------
    T : float
        Temperature of massive particle species in MeV.
    mu : float
        Chemical potential of massive particle species in MeV.
    m : float
        Mass of particle in MeV.
    g : float
        Degrees of freedom of massive particle species (1 for scalar, 
        3 for vector, etc.).

    Returns
    -------
    float
        Units of MeV^4. 
    """

    def res(i, val): 

        return val + jnp.exp(i * mu / T) * 3 * (m / T)**2 / i**2 * K2(i * m / T)

    return jnp.where(
        (T / m > rel_thres) | (T <= 0.), 
        p_massless_BE(T, mu, g), 
        g / (6 * jnp.pi**2) * T**4 * lax.fori_loop(1, N_series_terms, res, 0)
    )

def rho_massive_FD(T, mu, m, g):
    """
    Series approximation for energy density of a massive
    particle with Fermi-Dirac statistcs.

    Parameters
    ----------
    T : float
        Temperature of massive particle species in MeV.
    mu : float
        Chemical potential of massive particle species in MeV.
    m : float
        Mass of particle in MeV.
    g : float
        Degrees of freedom of massive particle species (2 for 
        Majorana Fermion, 4 for Dirac Fermion, etc.).

    Returns
    -------
    float
        Units of MeV^4. 
    """ 

    def res(i, val): 

        return val + (-1)**(i - 1) * jnp.exp(i * mu / T) * (
            (m / T)**3 / i * K1(i * m / T) 
            + 3 * (m / T)**2 / i**2 * K2(i * m / T)
        )


    return jnp.where(
        (T / m > rel_thres) | (T <= 0.), 
        rho_massless_FD(T, mu, g), 
        g / (2 * jnp.pi**2) * T**4 * lax.fori_loop(1, N_series_terms, res, 0)
    )

def n_massive_FD(T, mu, m, g): 
    """
    Series approximation for number density of a massive
    particle with Fermi-Dirac statistcs.

    Parameters
    ----------
    T : float
        Temperature of massive particle species in MeV.
    mu : float
        Chemical potential of massive particle species in MeV.
    m : float
        Mass of particle in MeV.
    g : float
        Degrees of freedom of massive particle species (2 for 
        Majorana Fermion, 4 for Dirac Fermion, etc.).

    Returns
    -------
    float
        Units of MeV^3. 
    """ 

    def res(i, val): 

        return val + (
            (-1)**(i - 1) * jnp.exp(i * mu / T) * (m / T)**2 / i * K2(i * m / T)
        )

    return jnp.where(
        (T / m > rel_thres) | (T <= 0.), 
        n_massless_FD(T, mu, g), 
        g / (2 * jnp.pi**2) * T**3 * lax.fori_loop(1, N_series_terms, res, 0)
    )

def p_massive_FD(T, mu, m, g): 
    """
    Series approximation for pressure of a massive
    particle with Fermi-Dirac statistcs.
    
    Parameters
    ----------
    T : float
        Temperature of massive particle species in MeV.
    mu : float
        Chemical potential of massive particle species in MeV.
    m : float
        Mass of particle in MeV.
    g : float
        Degrees of freedom of massive particle species (1 for 
        Majorana Fermion, 2 for Dirac Fermion, etc.).

    Returns
    -------
    float
        Units of MeV^4. 
    """ 

    def res(i, val): 

        return val + (
            (-1)**(i - 1) * jnp.exp(i * mu / T) 
            * 3 * (m / T)**2 / i**2 * K2(i * m / T)
        )

    return jnp.where(
        (T / m > rel_thres) | (T <= 0.), 
        p_massless_FD(T, mu, g), 
        g / (6 * jnp.pi**2) * T**4 * lax.fori_loop(1, N_series_terms, res, 0)
    )

def rho_massive_MB(T, mu, m, g): 
    """
    Series approximation for energy density of a massive
    particle with Maxwell-Boltzmann statistcs.

    Parameters
    ----------
    T : float
        Temperature of massive particle species in MeV.
    mu : float
        Chemical potential of massive particle species in MeV
    m : float
        Mass of particle in MeV.
    g : float
        Degrees of freedom of massive particle species.

    Returns
    -------
    float
        Units of MeV^4. 
    """ 

    return lax.cond(
        T > 0., 
        lambda T: g * m**2 * T * jnp.exp(mu/T) / (2 * jnp.pi**2) * (
            m * K1(m / T) + 3 * T * K2(m / T)
        ), 
        lambda T: 0., 
        T
    )

def n_massive_MB(T, mu, m, g): 
    """
    Series approximation for number density of a massive
    particle with Maxwell-Boltzmann statistcs.

    Parameters
    ----------
    T : float
        Temperature of massive particle species in MeV.
    mu : float
        Chemical potential of massive particle species in MeV.
    m : float
        Mass of particle in MeV.
    g : float
        Degrees of freedom of massive particle species.

    Returns
    -------
    float
        Units of MeV^3. 
    """ 

    return lax.cond(
        T > 0., 
        lambda T: g * m**2 * T * jnp.exp(mu/T) / (2 * jnp.pi**2) * K2(m / T), 
        lambda T: 0., 
        T
    )

def p_massive_MB(T, mu, m, g): 
    """
    Series approximation for pressure of a massive
    particle with Maxwell-Boltzmann statistcs.

    Parameters
    ----------
    T : float
        Temperature of massive particle species in units of MeV.
    mu : float
        Chemical potential of massive particle species.
    m : float
        Mass of particle in units of MeV.
    g : float
        Degrees of freedom of massive particle species.

    Returns
    -------
    float
        Units of MeV^4. 
    """ 

    return n_massive_MB(T, mu, m, g) * T

####################################################################
#               Electromagnetic Sector and Neutrinos               #
####################################################################

file_dir = os.path.dirname(__file__)

# QED Corrections
P_QED_tab = np.loadtxt(file_dir+"/data/background/"+"QED_P_int.txt")
dPdT_QED_tab = np.loadtxt(file_dir+"/data/background/"+"QED_dP_intdT.txt")
d2PdT2_QED_tab = np.loadtxt(file_dir+"/data/background/"+"QED_d2P_intdT2.txt")

# Effect of standard value of electron mass in scattering matrix elements
f_nue_scat_tab = np.loadtxt(file_dir+"/data/background/"+"nue_scatt.txt")
f_numu_scat_tab = np.loadtxt(file_dir+"/data/background/"+"numu_scatt.txt")

# Effect of standard value of electron mass in annihilation matrix elements
f_nue_ann_tab = np.loadtxt(file_dir+"/data/background/"+"nue_ann.txt")
f_numu_ann_tab = np.loadtxt(file_dir+"/data/background/"+"numu_ann.txt")

######################
# Standard EM Sector #
######################

def rho_EM_std(T_g, mu=0, LO=True, NLO=True): 
    """
    Total energy density of EM-coupled SM fluids.

    Parameters
    ----------
    T_g : float
        Photon temperature in MeV.
    mu : float, optional
        Parameter added for syntax consistency--does not impact function
        behavior.  Defaults to 0.
    LO : bool
        True includes leading order QED corrections to the energy density.
        Defaults to 'True'.
    NLO : bool
        True includes next-to-leading order QED corrections to the 
        energy density.  Defaults to 'True'.
        
    Returns
    -------
    float
        Units of MeV^4. 
    """ 

    corr_QED = (
        -jnp.interp(
            T_g, jnp.flip(P_QED_tab[:,0]), 
            jnp.flip(LO*P_QED_tab[:,1]+NLO*P_QED_tab[:,2])
        ) 
        + T_g*jnp.interp(
            T_g, jnp.flip(dPdT_QED_tab[:,0]),
            jnp.flip(LO*dPdT_QED_tab[:,1]+NLO*dPdT_QED_tab[:,2])
        )
    )

    return (
        rho_massless_BE(T_g, 0., 2)  + rho_massive_FD(T_g, 0., const.me, 4) 
        + corr_QED
    )

rho_EM_std_v = vmap(rho_EM_std, in_axes=0)

def p_EM_std(T_g, mu=0, LO=True, NLO=True): 
    """
    Total pressure of EM-coupled SM fluids.

    Parameters
    ----------
    T_g : float 
        Temperature of massive particle species in MeV.
    mu : float, optional
        Parameter added for syntax consistency--does not impact function
        behavior.  Defaults to 0.
    LO : bool
        True includes leading order QED corrections to the pressure.
        Defaults to 'True'.
    NLO : bool
        True includes next-to-leading order QED corrections to the 
        pressure.  Defaults to 'True'.
        
    Returns
    -------
    float 
        Units of MeV^4. 
    """ 

    corr_QED = jnp.interp(
        T_g, jnp.flip(P_QED_tab[:,0]),
        jnp.flip(LO*P_QED_tab[:,1] + NLO*P_QED_tab[:,2])
    )

    return (
        p_massless_BE(T_g, 0., 2) + p_massive_FD(T_g, 0., const.me, 4) 
        + corr_QED
    )

p_EM_std_v = vmap(p_EM_std, in_axes=0)

def rho_plus_p_EM_std(T_g, mu=0, LO=True, NLO=True): 
    """
    Sum of energy densities and pressures of all EM-coupled SM fluids.

    Parameters
    ----------
    T_g : float 
        Photon temperature in MeV.
    mu : float, optional
        Parameter added for syntax consistency--does not impact function
        behavior.  Defaults to 0.
    LO : bool
        True includes leading order QED corrections to the energy density 
        and pressure.  Defaults to 'True'.
    NLO : bool
        True includes next-to-leading order QED corrections to the 
        energy density and pressure.  Defaults to 'True'.
        
    Returns
    -------
    float 
        Units of MeV^4. 
    """ 

    corr_QED = T_g * jnp.interp(
        T_g, jnp.flip(dPdT_QED_tab[:,0]),
        jnp.flip(LO*dPdT_QED_tab[:,1] + NLO*dPdT_QED_tab[:,2])
    )
    
    return (
        4/3 * rho_massless_BE(T_g, 0., 2) + rho_massive_FD(T_g, 0., 0.511, 4) 
        + p_massive_FD(T_g, 0., 0.511, 4) + corr_QED
    )

def T_g(rho_g):
    """
    Photon temperature from photon energy density. 

    Parameters
    ----------
    rho_g : float or array
        Energy density of photons. 
    
    Returns
    -------
    float or array
        Same units as (rho_g)**0.25
    """
    return (30*rho_g/(2*jnp.pi**2))**(1/4)

drho_EM_dT_g_std = grad(rho_EM_std, argnums=0)

############################
# Standard Neutrino Sector #
############################ 

def rho_nue_std(T_nue, mu_nue=0.): 
    """
    Total energy density of electron neutrinos.

    Parameters
    ----------
    T_nue : float 
        Electron neutrino temperature in MeV.
    mu_nue : float, optional
        Chemical potential of electron neutrinos in MeV.  Defaults to 0.
        
    Returns
    -------
    float
        Units of MeV^4
    """ 

    return rho_massless_FD(T_nue, mu_nue, 2)

def p_nue_std(T_nue, mu_nue=0.): 
    """
    Total pressure of electron neutrinos.

    Parameters
    ----------
    T_nue : float
        Electron neutrino temperature in MeV.
    mu_nue : float, optional
        Chemical potential of electron neutrinos in MeV.  Defaults to 0.
        
    Returns
    -------
    float
        Units of MeV^4. 
    """ 

    return rho_massless_FD(T_nue, mu_nue, 2) / 3

def n_nue_std(T_nue, mu_nue=0.): 
    """
    Total number density of electron neutrinos.

    Parameters
    ----------
    T_nue : float
        Electron neutrino temperature in MeV.
    mu_nue : float, optional
        Chemical potential of electron neutrinos in MeV.  Defaults to 0.
        
    Returns
    -------
    float
        Units of MeV^3. 
    """ 

    return n_massless_FD(T_nue, mu_nue, 2) 

def rho_numt_std(T_numt, mu_numt=0.): 
    """
    Total energy density of mu, tau neutrinos.

    Parameters
    ----------
    T_nue : float
        Mu, tau neutrino temperature in MeV.
    mu_nue : float, optional
        Chemical potential of mu,tau neutrinos in MeV. Defaults to 0.
        
    Returns
    -------
    float
        Units of MeV^4. 
    """ 

    return 2 * rho_massless_FD(T_numt, mu_numt, 2)

def p_numt_std(T_numt, mu_numt=0.): 
    """
    Total pressure of mu, tau neutrinos.

    Parameters
    ----------
    T_nue : float
        Mu, tau neutrino temperature in MeV.
    mu_nue : float, optional
        Chemical potential of mu,tau neutrinos in MeV.  Defaults to 0.
        
    Returns
    -------
    float
        Units of MeV^4. 
    """ 

    return 2 * rho_massless_FD(T_numt, mu_numt, 2) / 3

def n_numt_std(T_numt, mu_numt=0.): 
    """
    Total number density of mu, tau neutrinos.

    Parameters
    ----------
    T_nue : float
        Mu, tau neutrino temperature in MeV.
    mu_nue : float, optional
        Chemical potential of mu,tau neutrinos in MeV.  Defaults to 0.
        
    Returns
    -------
    float
        Units of MeV^4. 
    """ 

    return 2 * n_massless_FD(T_numt, mu_numt, 2)

def T_nu(rho_nu):
    """
    Neutrino temperature given an energy density.

    Parameters
    ----------
    rho_nu : float or array
        Neutrino energy density in MeV^4.
        
    Returns
    -------
    float or array
        Units of MeV. 
    """ 
    return (8./7.*30*rho_nu/(2*jnp.pi**2))**(1/4)

drho_nue_dT_nue_std  = grad(rho_nue_std, argnums=0)
drho_nue_dmu_nue_std = grad(rho_nue_std, argnums=1)
dn_nue_dT_nue_std    = grad(n_nue_std, argnums=0)
dn_nue_dmu_nue_std   = grad(n_nue_std, argnums=1)

drho_numt_dT_numt_std  = grad(rho_numt_std, argnums=0)
drho_numt_dmu_numt_std = grad(rho_numt_std, argnums=1)
dn_numt_dT_numt_std    = grad(n_numt_std, argnums=0)
dn_numt_dmu_numt_std   = grad(n_numt_std, argnums=1)


def collision_terms_std(
    T_g, T_nue, T_numt, mu_nue=0., mu_numt=0., 
    decoupled=False, use_FD=True, collision_me=True
): 
    """
    Energy and number density transfer rate between EM and neutrino sector 
    relevant for incomplete neutrino decoupling.

    Parameters
    ----------
    T_g : array_like 
        Photon temperature in MeV.
    T_nue : array_like
        Electron neutrino temperature in MeV.
    T_numt : array_like
        Mu, tau neutrino temperature in MeV.
    mu_nue : float, optional
        Chemical potential of electron neutrinos in MeV.
        Defaults to 0.
    mu_numt : float, optional
        Chemical potential of mu, tau neutrinos in MeV.
        Defaults to 0.
    decoupled : bool, optional
        Neutrinos are assumed to be completely decoupled if True. 
    use_FD : bool, optional
        Fermi-Dirac distribution used for neutrinos if True. 
    collision_me : bool, optional
        Finite electron mass if true. 
        
    Returns
    -------
    tuple
        (C_rho_nue, C_rho_numu, C_n_nue, C_n_numu) for the energy density 
        transfer rate (in MeV^4/s) to nu_e and (nu_mu, nu_tau), followed by 
        the number density transfer rate (in MeV^3/s) to nu_e and (nu_mu, 
        nu_tau)

    
    """ 

    f_n, f_a, f_s = lax.cond(
        decoupled, 
        lambda _: (0., 0., 0.), 
        lambda _: lax.cond(
            use_FD, lambda _: (0.852, 0.884, 0.829), 
            lambda _: (1., 1., 1.), 0.
        ), 
        0.
    )

    def G(T_1, mu_1, T_2, mu_2): 
        
        return (
            32 * f_a * (
                T_1**9 * jnp.exp(2 * mu_1 / T_1) 
                - T_2**9 * jnp.exp(2 * mu_2 / T_2)
            ) 
            + 56 * f_s * jnp.exp(2 * mu_1 / T_1) * jnp.exp(2 * mu_2 / T_2) *(   
                T_1**4 * T_2**4 * (T_1 - T_2)
            )
        )

    def G_nue_with_me(T_1, mu_1, T_2, mu_2):

        def interp_f(f_tab): 

            return jnp.interp(
                T_1, f_tab[:,0], f_tab[:,1], left=f_tab[0,1], right=f_tab[-1,1]
            )

        f_nue_ann  = lax.cond(
            collision_me, interp_f, lambda _: 1., f_nue_ann_tab
        )
        f_nue_scat = lax.cond(
            collision_me, interp_f, lambda _: 1., f_nue_scat_tab
        )
        
        return (
            32 * f_a * f_nue_ann * (
                T_1**9 * jnp.exp(2 * mu_1 / T_1) 
                - T_2**9 * jnp.exp(2 * mu_2 / T_2)
            ) 
            + 56 * f_s * f_nue_scat * (
                jnp.exp(2 * mu_1 / T_1) * jnp.exp(2 * mu_2 / T_2) 
                * T_1**4 * T_2**4 * (T_1 - T_2)
            )
        )

    def G_numt_with_me(T_1, mu_1, T_2, mu_2): 

        def interp_f(f_tab): 

            return jnp.interp(
                T_1, f_tab[:,0], f_tab[:,1], left=f_tab[0,1], right=f_tab[-1,1]
            )

        f_numt_ann  = lax.cond(
            collision_me, interp_f, lambda _: 1., f_numu_ann_tab
        )
        f_numt_scat = lax.cond(
            collision_me, interp_f, lambda _: 1., f_numu_scat_tab
        )
        
        return (
            32 * f_a * f_numt_ann * (
                T_1**9 * jnp.exp(2 * mu_1 / T_1) 
                - T_2**9 * jnp.exp(2 * mu_2 / T_2)
            ) 
            + 56 * f_s * f_numt_scat * (
                jnp.exp(2 * mu_1 / T_1) * jnp.exp(2 * mu_2 / T_2) 
                * T_1**4 * T_2**4 * (T_1 - T_2)
            )
        )

    geL  = const.geL
    geR  = const.geR
    gmuL = const.gmuL
    gmuR = const.gmuR

    # Units MeV^4 s^-1
    C_rho_nue = const.GF**2 / jnp.pi**5 * (
        4 * (geL**2 + geR**2) * G_nue_with_me(T_g, 0., T_nue, mu_nue) 
        + 2 * G(T_numt, mu_numt, T_nue, mu_nue)
    ) / const.hbar

    # Units MeV^4 s^-1
    C_rho_numu = const.GF**2 / jnp.pi**5 * (
        4 * (gmuL**2 + gmuR**2) * G_numt_with_me(T_g, 0., T_numt, mu_numt) 
        - G(T_numt, mu_numt, T_nue, mu_nue)
    ) / const.hbar 

    # Units MeV^3 s^-1
    C_n_nue = 8 * f_n * const.GF**2 / jnp.pi**5 * (
        4 * (geL**2 + geR**2) 
        * (T_g**8 - T_nue**8 * jnp.exp(2 * mu_nue / T_nue))
        + 2 * (
            T_numt**8 * jnp.exp(2 * mu_numt / T_numt) 
            - T_nue**8 * jnp.exp(2 * mu_nue / T_nue)
        )
    ) / const.hbar

    # Units MeV^3 s^-1
    C_n_numu = 8 * f_n * const.GF**2 / jnp.pi**5 * (
        4 * (gmuL**2 + gmuR**2) 
        * (T_g**8 - T_nue**8 * jnp.exp(2 * mu_numt / T_numt))
        - (
            T_numt**8 * jnp.exp(2 * mu_numt / T_numt) 
            - T_nue**8 * jnp.exp(2 * mu_nue / T_nue)
        )
    ) / const.hbar

    return C_rho_nue, C_rho_numu, C_n_nue, C_n_numu


#import abc
from jax import config, lax
import jax.numpy as jnp
import equinox as eqx
from . import constants as cnst

config.update("jax_enable_x64", True)

### ABSTRACT BASE CLASSES AND INTERFACES ###

class Fluid(eqx.Module):
    """
    Base class for fluid species.

    Defines fluid properties.

    Fields:
    -------
    delta_idx : int
        Default = 0
        Position of the first perturbation equation in the Diffrax vector. For most fluids this is the density perturbation mode "delta".
    num_moments : int 
        Default = 0
        Number of equations that need to be simultaneously evolved in the perturbations module. 
    name : str
        Default = ""
        Name of the fluid, used to find fluid and refer to it later in the computation using species_dict["name"].
    is_matter : bool
        Default = False
        Whether the fluid is non-relativistic today and contributes towards the total matter power spectrum. 

    Methods:
    --------
        rho : Compute energy density (units: eV cm^{-3})
        P   : Compute pressure (units: eV cm^{-3})
        w   : Compute equation of state parameter (units: dimensionless)
        y_ini   : Adiabatic initial conditions, in synchronous gauge
        y_prime : Perturbation derivatives, in synchronous gauge
        rho_delta        : Perturbed density function δρ (units: eV cm^{-3})
        rho_plus_P_theta : Velocity perturbation  (units: eV cm^{-3} Mpc^{-1})
        rho_plus_P_sigma : Compute standard shear perturbation (units: eV cm^{-3})
    """

    delta_idx     : int = eqx.field(default=0)
    num_moments : int = eqx.field(default=0, static=True)
    name          : str = eqx.field(default="", static=True)
    is_matter     : bool = eqx.field(default=False) # Does the fluid contribute towards matter overdensity today.

    def __init__(self, delta_idx, specs):
        self.delta_idx = delta_idx
        self.name = ""
        self.is_matter = False

    def rho(self, lna, args):
        """
        Compute energy density.

        Calculates the energy density of the fluid species at a given
        cosmological epoch using the logarithm of the scale factor.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Energy density (units: eV cm^{-3})
        """
        raise NotImplementedError("Fluid species must implement an energy density function.")

    def P(self, lna, args):
        """
        Compute pressure.

        Calculates the pressure of the fluid species at a given
        cosmological epoch using the logarithm of the scale factor.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Pressure (units: eV cm^{-3})
        """
        raise NotImplementedError("Fluid species must implement a pressure function.")

    def w(self, lna, args):
        """
        Compute equation of state parameter.

        Calculates the ratio of pressure to energy density, representing
        the equation of state for the fluid species.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Equation of state parameter (units: dimensionless)
        """
        return self.P(lna, args)/self.rho(lna, args)

    def y_ini(self, k, tau_ini, args):
        """
        Compute initial conditions for perturbation modes.

        Calculates the initial state of perturbation modes at early cosmological times.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        tau_ini : float
            Initial conformal time (units: Mpc)
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        array
            Initial perturbation mode values
        """
        raise NotImplementedError("Fluid species must implement the initial conditions of their perturbation modes.")

    def y_prime(self, k, lna, metric_h_prime, metric_eta_prime, y, args):
        """
        Compute time derivatives of perturbation modes.

        Calculates how perturbation modes evolve with cosmological time.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        lna : float
            Logarithm of scale factor
        metric_h_prime : float
            Derivative of metric h
        metric_eta_prime : float
            Derivative of metric eta
        y : array
            Current perturbation mode values
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        array
            Time derivatives of perturbation modes
        """
        raise NotImplementedError("Fluid species must implement a perturbation derivative function.")

    def rho_delta(self, lna, y, args):
        """
        Compute density perturbation.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : array
            Perturbation mode values
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Density perturbation (units: eV cm^{-3})
        """
        raise NotImplementedError("Fluid species must implement a perturbation derivative function.")

    def rho_plus_P_theta(self, lna, y, args):
        """
        Compute velocity perturbation.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : array
            Perturbation mode values
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Velocity perturbation (units: eV cm^{-3} Mpc^{-1})
        """
        raise NotImplementedError("Fluid species must implement a perturbation derivative function.")

    def rho_plus_P_sigma(self, lna, y, args):
        """
        Compute shear perturbation.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : array
            Perturbation mode values
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Shear perturbation (units: eV cm^{-3})
        """
        raise NotImplementedError("Fluid species must implement a perturbation derivative function.")

class StandardFluid(Fluid):
    """
    Standard implementation of perturbation methods for fluid species.

    Provides default computations for perturbation-related methods
    used in this code.

    Methods:
    --------
    rho_delta : Compute standard density perturbation (units: eV cm^{-3})
    rho_plus_P_theta : Compute standard velocity perturbation (units: eV cm^{-3} Mpc^{-1})
    rho_plus_P_sigma : Compute standard shear perturbation (units: eV cm^{-3})
    """

    def __init__(self, delta_idx, specs):
        super().__init__(delta_idx, specs)

    # Called by diffrax, child classes should never override. Okay to implement here.
    def rho_delta(self, lna, y, args):
        """
        Compute density perturbation.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : array
            Perturbation mode values
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Density perturbation (units: eV cm^{-3})
        """
        params = args
        return self.rho(lna, params) * y[self.delta_idx]

    def rho_plus_P_theta(self, lna, y, args):
        """
        Compute velocity perturbation.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : array
            Perturbation mode values
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Velocity perturbation (units: eV cm^{-3} Mpc^{-1})
        """
        params = args
        return jnp.where(
            self.num_moments > 1,
            (self.rho(lna, params)+self.P(lna, params)) * y[self.delta_idx+1],
            0.
        )

    def rho_plus_P_sigma(self, lna, y, args):
        """
        Compute shear perturbation.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : array
            Perturbation mode values
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Shear perturbation (units: eV cm^{-3})
        """
        params = args
        return jnp.where(
            self.num_moments > 2,
            (self.rho(lna, params)+self.P(lna, params)) * y[self.delta_idx+2],
            0.
        )

class BackgroundFluid(Fluid):
    
    num_moments = 0

    def __init__(self, delta_idx, specs):
        super().__init__(delta_idx, specs)

    def y_ini(self, k, tau_ini, args):
        """
        Trivial initial condition vector for background.
        """
        return jnp.array([])

    def y_prime(self, k, lna, metric_h_prime, metric_eta_prime, y, args):
        """
        Trivial derivative vector for background. 
        """
        return jnp.array([])

    def rho_delta(self, lna, y, args):
        return 0.

    def rho_plus_P_theta(self, lna, y, args):
        return 0.

    def rho_plus_P_sigma(self, lna, y, args):
        return 0.

### BEGINNING OF CONCRETE CLASSES ###

class DarkEnergy(BackgroundFluid):
    """
    Dark energy fluid species implementation.

    Represents a constant energy density fluid with negative pressure.

    Required input parameters: None.

    Required derived parameters: params['omega_Lambda']

    Methods:
    --------
    rho : Compute dark energy density (units: eV cm^{-3})
    P : Compute dark energy pressure (units: eV cm^{-3})
    """

    name = "DarkEnergy"

    def __init__(self, delta_idx, specs):
        super().__init__(delta_idx, specs)
        self.name = "DarkEnergy"

    def rho(self, lna, args):
        """
        Compute dark energy density.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Dark energy density (units: eV cm^{-3})
        """
        params = args
        return params['omega_Lambda'] * (3.*cnst.H0_over_h**2/8./jnp.pi/cnst.G)
    
    def P(self, lna, args):
        """
        Compute dark energy pressure.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Dark energy pressure (units: eV cm^{-3})
        """
        params = args
        return -self.rho(lna, params)

class ColdDarkMatter(StandardFluid):
    """
    Cold dark matter fluid species implementation.

    Non-relativistic, pressureless dark matter with density
    perturbations but no velocity or shear modes.

    Required input parameters: params['omega_cdm'].

    Required derived parameters: params['om'].

    Methods:
    --------
    rho : Compute cold dark matter density (units: eV cm^{-3})
    P : Compute cold dark matter pressure (units: eV cm^{-3})
    y_ini : Compute initial perturbation conditions
    y_prime : Compute perturbation time derivatives
    """

    name = "ColdDarkMatter"
    num_moments = 1  # CDM only receives density perturbation in synchronous gauge.
    is_matter = True

    def __init__(self, delta_idx, specs):
        super().__init__(delta_idx, specs)
        self.name = "ColdDarkMatter"
        self.is_matter = True

    def rho(self, lna, args):
        """
        Compute cold dark matter density.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Cold dark matter density (units: eV cm^{-3})
        """
        params = args
        return params['omega_cdm'] * (3.*cnst.H0_over_h**2/8./jnp.pi/cnst.G) / jnp.exp(lna)**3

    def P(self, lna, args):
        """
        Compute cold dark matter pressure.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Cold dark matter pressure (units: eV cm^{-3})

        Notes:
        ------
        Cold dark matter is pressureless, so this always returns zero.
        """
        return 0.
    
    def y_ini(self, k, tau_ini, args):
        """
        Compute initial conditions for cold dark matter perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        tau_ini : float
            Initial conformal time (units: Mpc)
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        array
            Initial density perturbation (units: dimensionless)
        """
        params = args
        delta = -(k*tau_ini)**2/4. * (1.-params["om"]*tau_ini/5.)
        return jnp.array([delta])

    def y_prime(self, k, lna, metric_h_prime, metric_eta_prime, y, args):
        """
        Compute time derivatives of cold dark matter perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        lna : float
            Logarithm of scale factor
        metric_h_prime : float
            Derivative of metric h
        metric_eta_prime : float
            Derivative of metric eta
        y : array
            Current perturbation mode values
        args : tuple
            Background cosmology and cosmological parameters (BG, params) - Note: BG parameter is unused in this implementation

        Returns:
        --------
        array
            Time derivative of density perturbation (units: dimensionless)
        """
        return jnp.array([-0.5*metric_h_prime])

class MasslessNeutrino(StandardFluid):
    """
    Massless neutrinos fluid species implementation.

    Represents relativistic neutrinos with multiple angular momentum modes.

    Required input parameters: params['N_nu_massless'], params['T_nu_massless'], params['TCMB0']

    Required derived parameters: params['R_nu'], params['om']

    Methods:
    --------
    rho : Compute neutrino density (units: eV cm^{-3})
    P : Compute neutrino pressure (units: eV cm^{-3})
    """

    name = "MasslessNeutrino"

    def __init__(self, delta_idx, specs):
        super().__init__(delta_idx, specs)
        self.name = "MasslessNeutrino"
        self.num_moments = specs["l_max_massless_nu"] + 1

    def rho(self, lna, args):
        """
        Compute neutrino density.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Neutrino density (units: eV cm^{-3})
        """
        params = args

        a = jnp.exp(lna)
        rho = params['N_nu_massless'] * 2. * 7./8. * jnp.pi**2/30. * params['T_nu_massless']**4 * params['TCMB0']**4 / a**4 # eV^4
        rho = rho / (cnst.c * cnst.hbar)**3 # Convert to eV cm^{-3}
        return rho
    
    def P(self, lna, args):
        """
        Compute neutrino pressure.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Neutrino pressure (units: eV cm^{-3})
        """
        params = args
        return self.rho(lna, params)/3.

    def y_ini(self, k, tau_ini, args):
        """
        Compute initial conditions for massless neutrino perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        tau_ini : float
            Initial conformal time (units: Mpc)
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        array
            Initial perturbation mode values (units: 1/Mpc for theta, else dimensionless)
        """
        params = args
        R_nu = params['R_nu']

        delta = - (k*tau_ini)**2/3. * (1.-params["om"]*tau_ini/5.)
        theta = - k*(k*tau_ini)**3/36./(4.*R_nu+15.) \
                * (4.*R_nu+11.+12.-3.*(8.*R_nu**2+50.*R_nu+275.)/20./(2.*R_nu+15.)*tau_ini*params["om"])
        sigma = (k*tau_ini)**2/(45.+12.*R_nu) * 2. * (1.+(4.*R_nu-5.)/4./(2.*R_nu+15.)*tau_ini*params["om"])
        
        # Return the four non-zero ell modes, and all higher ell-modes are zero to start.
        # For the neutrinos we track Fnu_2 = 2*sigma, for better structure within the hierarchy.
        return jnp.concatenate((jnp.array([delta, theta, sigma]), jnp.zeros(self.num_moments-3)))

    def y_prime(self, k, lna, metric_h_prime, metric_eta_prime, y, args):
        """
        Compute time derivatives of massless neutrino perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        lna : float
            Logarithm of scale factor
        metric_h_prime : float
            Derivative of metric h
        metric_eta_prime : float
            Derivative of metric eta
        y : array
            Current perturbation mode values
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        array
            Time derivatives of perturbation modes (units: 1/Mpc for theta, else dimensionless)
        """
        BG, params, _, _ = args
        aH    = BG.aH(lna, params)
        tau   = BG.tau(lna)

        L = jnp.arange(self.num_moments) + self.delta_idx
        F = y[L]
        delta = F[0]
        theta = F[1]
        sigma = F[2]

        # density, velocity, shear perturbations
        delta_prime = -4./3./aH*theta - 2./3.*metric_h_prime
        theta_prime = k**2/aH*(delta/4.-sigma)
        sigma_prime = 4./15./aH*theta - 3./10.*k/aH*F[3] + 2./15.*metric_h_prime + 4./5.*metric_eta_prime
        F3_prime = 1./7. * k/aH * (6.*sigma - 4.*F[4])

        # Rest of the Boltzmann Hierarchy
        lmax = self.num_moments-1
        L = jnp.arange(4, lmax)
        Fl_prime    = 1./(2.*L+1.)*k/aH * (L*F[L-1]-(L+1)*F[L+1])
        Flmax_prime = k/aH*F[lmax-1] - (lmax+1)/aH/tau*F[lmax]

        return jnp.concatenate((jnp.array([delta_prime, theta_prime, sigma_prime, F3_prime]), Fl_prime, jnp.array([Flmax_prime])))

class MassiveNeutrino(Fluid):
    """
    Massive neutrinos fluid species implementation.

    Non-relativistic neutrinos with multiple angular momentum modes.

    Required input parameters: params['N_nu_massive'], params['T_nu_massive'], 
    params['m_nu_massive'], params['TCMB0']

    Required derived parameters: params['R_nu'], params['om']

    Attributes:
    -----------
    num_ells_per_bin : int
        Number of multipole moments per momentum bin for massive neutrino hierarchy

    Methods:
    --------
    rho : Compute massive neutrino density (units: eV cm^{-3})
    P : Compute massive neutrino pressure (units: eV cm^{-3})
    y_ini : Compute initial perturbation conditions 
    y_prime : Compute perturbation time derivatives 
    rho_delta : Compute density perturbation (units: eV cm^{-3})
    rho_plus_P_theta : Compute velocity perturbation (units: eV cm^{-3} Mpc^{-1})
    rho_plus_P_sigma : Compute shear perturbation (units: eV cm^{-3})
    """

    num_ells_per_bin : int = eqx.field(static=True)

    q_3p = jnp.array([0.913201, 3.37517, 7.79184])
    w_3p = jnp.array([0.0687359, 3.31435, 2.29911])
    q_5p = jnp.array([0.583165, 2.0, 4.0, 7.26582, 13.0])
    w_5p = jnp.array([0.0081201, 0.689407, 2.8063, 2.05156, 0.12681])

    dlfdlq_3p = -q_3p / (1.+jnp.exp(-q_3p)) # Log derivative of fermi-dirac w.r.t. momentum

    name = "MassiveNeutrino"
    is_matter = True

    def __init__(self, delta_idx, specs):

        super().__init__(delta_idx, specs)
        self.name = "MassiveNeutrino"
        self.is_matter = True
        self.num_ells_per_bin = specs["l_max_massive_nu"] + 1
        self.num_moments = 3 * self.num_ells_per_bin

    def rho(self, lna, args):
        """
        Compute massive neutrino density.

        Parameters:
        -----------
        lna : float or ArrayLike
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float or ArrayLike
            Massive neutrino density (units: eV cm^{-3})
        """
        params = args

        # Ensure lna is at least 1D for broadcasting
        lna_arr = jnp.atleast_1d(lna)  # shape (N,)
        # shape (N,1):
        a = jnp.exp(lna_arr)[:, None]         
        T = params['T_nu_massive'] * params['TCMB0'] / a  
        x = params['m_nu_massive'] / T 

        # q_5p, w_5p are shape (5,), broadcast with (N, 1)
        integrand = (1. + jnp.exp(-self.q_5p)) / self.q_5p**2 \
                    * jnp.sqrt(self.q_5p**2 + x**2)           # (N, 5)

        # Dot product along last axis with w_5p
        integral = jnp.dot(integrand, self.w_5p)               # (N,)

        rho_val = params['N_nu_massive'] * 4. * T[:, 0]**4 / jnp.pi**2 * integral / cnst.hbar**3 / cnst.c**3

        # Remove extra dimension if original input was scalar
        return jnp.squeeze(rho_val) if jnp.ndim(lna) == 0 else rho_val

    def P(self, lna, args):
        """
        Compute massive neutrino pressure.

        Parameters:
        -----------
        lna : float or ArrayLike
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float or ArrayLike
            Massive neutrino pressure (units: eV cm^{-3})
        """
        params = args

        # Ensure lna is at least 1D for broadcasting
        lna_arr = jnp.atleast_1d(lna)          # shape (N,)
        # shape (N,1)
        a = jnp.exp(lna_arr)[:, None] 
        T = params['T_nu_massive'] * params['TCMB0'] / a 
        x = params['m_nu_massive'] / T 

        # q_5p, w_5p are shape (5,), broadcast with (N, 1)
        integrand = (1. + jnp.exp(-self.q_5p)) / jnp.sqrt(self.q_5p**2 + x**2) # (N, 5)

        # Dot product along last axis with w_5p
        integral = jnp.dot(integrand, self.w_5p)               # (N,)

        P_val = params['N_nu_massive'] * 4./3. * T[:, 0]**4 / jnp.pi**2 * integral / cnst.hbar**3 / cnst.c**3

        # Remove extra dimension if original input was scalar
        return jnp.squeeze(P_val) if jnp.ndim(lna) == 0 else P_val

    def y_ini(self, k, tau_ini, args):
        """
        Compute initial conditions for massive neutrino perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        tau_ini : float
            Initial conformal time (units: Mpc)
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        array
            Initial perturbation mode values (units: 1/Mpc for kPsi1, else dimensionless)
        """
        params = args
        res = jnp.zeros(self.num_moments)

        # Initial conditions for massless neutrinos first, needed here.
        R_nu = params['R_nu']

        delta = - (k*tau_ini)**2/3. * (1.-params["om"]*tau_ini/5.)
        theta = - k*(k*tau_ini)**3/36./(4.*R_nu+15.) \
                * (4.*R_nu+11.+12.-3.*(8.*R_nu**2+50.*R_nu+275.)/20./(2.*R_nu+15.)*tau_ini*params["om"])
        sigma = (k*tau_ini)**2/(45.+12.*R_nu) * 2. * (1.+(4.*R_nu-5.)/4./(2.*R_nu+15.)*tau_ini*params["om"])

        idx_q1 = 0 # Psi0 index of first q
        idx_q2 = idx_q1 + self.num_ells_per_bin # Psi0 index of second q
        idx_q3 = idx_q2 + self.num_ells_per_bin # Psi0 index of third q

        for i in range(3):
            q  = self.q_3p[i] # This momentum bin.
            iq = i*self.num_ells_per_bin # Index in diffrax array of this momentum bin.
            # ZZ : Techniclly Psi1 requires epsilon/q = 1/v, but at early times this should be 1. Should check this accuracy!
            first_three = jnp.array([delta/4., theta/3., sigma/2.]) * q / (1.+jnp.exp(-q))
            res = res.at[iq:iq+3].set(first_three)

        return res

    def y_prime(self, k, lna, metric_h_prime, metric_eta_prime, y, args):
        """
        Compute time derivatives of massive neutrino perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        lna : float
            Logarithm of scale factor
        metric_h_prime : float
            Derivative of metric h
        metric_eta_prime : float
            Derivative of metric eta
        y : array
            Current perturbation mode values
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        array
            Time derivatives of perturbation modes (units: 1/Mpc for kPsi1, else dimensionless)
        """
        BG, params, _, _ = args
        res = jnp.zeros(self.num_moments)

        a = jnp.exp(lna)
        T = params['T_nu_massive'] * params['TCMB0'] / a
        x = params['m_nu_massive'] / T
        aH  = BG.aH(lna, params)
        tau = BG.tau(lna)

        # Iterate through momentum bins
        for i in range(3):
            q = self.q_3p[i]
            epsilon = jnp.sqrt(q**2 + x**2)
            dlnf0_dlnq = -q / (1+jnp.exp(-q))

            # NOTE: The entries are [Psi0, k * Psi1, Psi2, ...]. If accessing Psi1 make sure to divide out k
            L = jnp.arange(self.num_ells_per_bin) + self.delta_idx + i*self.num_ells_per_bin
            Psi = y[L]

            Psi0_prime = -q/epsilon/aH*Psi[1] + metric_h_prime/6. * dlnf0_dlnq
            kPsi1_prime = q*k**2/3./epsilon/aH * (Psi[0] - 2.*Psi[2])
            Psi2_prime = q*k/5./epsilon/aH * (2.*Psi[1]/k - 3.*Psi[3]) - (metric_h_prime/15. + 2.*metric_eta_prime/5.) * dlnf0_dlnq
            
            # Intermediate hierarchy, 3<=L<lmax
            lmax = self.num_ells_per_bin - 1
            L_inter = jnp.arange(3, lmax) # Doesn't include lmax.
            Psi_inter_prime = q*k/epsilon/aH/(2*L_inter+1) * (L_inter*Psi[L_inter-1] - (L_inter+1)*Psi[L_inter+1])

            # lmax mode
            Psi_lmax_prime = q*k/aH/epsilon*Psi[lmax-1] - (lmax+1)/aH/tau*Psi[lmax]

            # Putting it all together
            subres = jnp.concatenate((jnp.array([Psi0_prime, kPsi1_prime, Psi2_prime]), Psi_inter_prime, jnp.array([Psi_lmax_prime])))
            res = res.at[i*self.num_ells_per_bin:(i+1)*self.num_ells_per_bin].set(subres)

        return res

    def rho_delta(self, lna, y, args):
        """
        Compute massive neutrino density perturbation.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : array
            Perturbation mode values
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Density perturbation (units: eV cm^{-3})
        """
        params = args
        a = jnp.exp(lna)
        T = params['T_nu_massive'] * params['TCMB0'] / a  # (N,)
        x = params['m_nu_massive'] / T  # (N,)

        res = 0.
        for i in range(3):
            q = self.q_3p[i]
            w = self.w_3p[i]
            epsilon = jnp.sqrt(q**2 + x**2)
            Psi0 = y[self.delta_idx + i*self.num_ells_per_bin]

            res += w*(1.+jnp.exp(-q))*epsilon/q**2 * Psi0
        return params['N_nu_massive'] * res * 4./jnp.pi**2 * T**4 / cnst.hbar**3 / cnst.c**3

    def rho_plus_P_theta(self, lna, y, args):
        """
        Compute massive neutrino velocity perturbation.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : array
            Perturbation mode values
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Velocity perturbation (units: eV cm^{-3} Mpc^{-1})
        """
        params = args
        a = jnp.exp(lna)
        T = params['T_nu_massive'] * params['TCMB0'] / a  # (N,)
        x = params['m_nu_massive'] / T  # (N,)

        res = 0.
        for i in range(3):
            q = self.q_3p[i]
            w = self.w_3p[i]
            kPsi1 = y[self.delta_idx+1 + i*self.num_ells_per_bin]

            res += w*(1.+jnp.exp(-q))/q * kPsi1
        return params['N_nu_massive'] * res * 4./jnp.pi**2 * T**4 / cnst.hbar**3 / cnst.c**3

    def rho_plus_P_sigma(self, lna, y, args):
        """
        Compute massive neutrino shear perturbation.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : array
            Perturbation mode values
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Shear perturbation (units: eV cm^{-3})
        """
        params = args
        a = jnp.exp(lna)
        T = params['T_nu_massive'] * params['TCMB0'] / a  # (N,)
        x = params['m_nu_massive'] / T  # (N,)

        res = 0.
        for i in range(3):
            q = self.q_3p[i]
            w = self.w_3p[i]
            epsilon = jnp.sqrt(q**2 + x**2)
            Psi2 = y[self.delta_idx+2 + i*self.num_ells_per_bin]

            res += w*(1.+jnp.exp(-q))/epsilon * Psi2
        return params['N_nu_massive'] * res * 8./3./jnp.pi**2 * T**4 / cnst.hbar**3 / cnst.c**3


class Baryon(StandardFluid):
    """
    Baryon fluid species implementation.

    Non-relativistic baryons with density and velocity perturbations.

    Required input parameters: params['omega_b'], params['YHe']

    Required derived parameters: params['R_nu'], params['R_b'], params['om']

    Methods:
    --------
    rho : Compute baryon density (units: eV cm^{-3})
    P : Compute baryon pressure (units: eV cm^{-3})
    cs2 : Compute sound speed squared (units: dimensionless)
    mean_mass : Compute mean baryon mass (units: eV)
    y_ini : Compute initial perturbation conditions 
    y_prime : Compute perturbation time derivatives 
    """
    
    name = "Baryon"
    num_moments = 2
    is_matter = True

    def __init__(self, delta_idx, specs):
        super().__init__(delta_idx, specs)
        self.name = "Baryon"
        self.is_matter = True

    def rho(self, lna, args):
        """
        Compute baryon density.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Baryon density (units: eV cm^{-3})
        """
        params = args
        return params['omega_b'] * (3.*cnst.H0_over_h**2/8./jnp.pi/cnst.G) / jnp.exp(lna)**3

    def P(self, lna, args):
        """
        Compute baryon pressure.

        Parameters:
        ------------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Baryon pressure (units: eV cm^{-3})

        Notes:
        ------
        Baryon pressure is neglected, standard practice for SM baryons.
        """
        return 0.

    def cs2(self, lna, args):
        """
        Compute sound speed squared.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        float
            Sound speed squared (units: dimensionless)

        Notes:
        ------
        Adiabatic sound speed squared, taken from M&B Eq. (68).
        Although we can neglect the pressure, this term is important for perturbation growth
        during recombination. During reionization this cs2 is negative. This is not physical
        but it should not matter for cosmology.
        """
        BG, params, species_list, species_dict = args
        # Get photon class from list
        i = species_dict["Photon"]
        photon = species_list[i]

        Tm = BG.Tm(lna, params) # Baryon temp
        Tg = BG.TCMB(lna, params) # Photon temp
        mu = self.mean_mass(lna, (BG,params))
        R = 4.*photon.rho(lna, params)/3./self.rho(lna, params)

        return Tm/mu * (5./3. - 2./3.*mu*R/cnst.me/BG.aH(lna, params)/BG.tau_c(lna, params) * (Tg/Tm - 1.))

    def mean_mass(self, lna, args):
        """
        Compute mean baryon mass at given redshift.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        float
            Mean baryon mass (units: eV)

        Notes:
        ------
        Defined to be mu = rho_b / n_b = rho_b / (nH + nHe + ne)
        """
        BG, params = args
        denom = (1.+BG.xe(lna))*(1.-params['YHe']) + cnst.mH / cnst.mHe * params['YHe']
        return cnst.mH / denom

    def y_ini(self, k, tau_ini, args):
        """
        Compute initial conditions for baryon perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        tau_ini : float
            Initial conformal time (units: Mpc)
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        array
            Initial perturbation mode values (units: 1/Mpc for theta, else dimensionless)
        """
        params = args
        delta = -(k*tau_ini)**2/4. * (1.-params["om"]*tau_ini/5.)
        theta = - k**4 * tau_ini**3/36. * (1.-3.*(1.+5.*params['R_b']-params['R_nu'])/20./(1.-params['R_nu'])*params["om"]*tau_ini)
        return jnp.array([delta, theta])

    def y_prime(self, k, lna, metric_h_prime, metric_eta_prime, y, args):
        """
        Compute time derivatives of baryon perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        lna : float
            Logarithm of scale factor
        metric_h_prime : float
            Derivative of metric h
        metric_eta_prime : float
            Derivative of metric eta
        y : array
            Current perturbation mode values
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        array
            Time derivatives of perturbation modes (units: 1/Mpc for theta, else dimensionless)
        """
        BG, params, species_list, species_dict = args
        # Get photon class from list
        i = species_dict["Photon"]
        photon = species_list[i]

        aH = BG.aH(lna, params)
        cs2 = self.cs2(lna, args)
        R = 4.*photon.rho(lna, params)/3./self.rho(lna, params)
        tau_c = BG.tau_c(lna, params)

        delta = y[self.delta_idx]
        theta = y[self.delta_idx+1]
        theta_g = y[photon.delta_idx+1]
        delta_prime = -theta/aH-metric_h_prime/2.
        theta_prime = -theta + cs2*k**2*delta/aH + R/tau_c/aH*(theta_g-theta)
        
        return jnp.array([delta_prime, theta_prime])

class Photon(StandardFluid):
    """
    Photon fluid species implementation.

    Relativistic photons with temperature and polarization Boltzmann hierarchies.

    Required input parameters: params['TCMB0']

    Required derived parameters: params['R_nu'], params['R_nu'], params['om']

    Attributes:
    -----------
    num_F_ell_modes : int
        Number of temperature multipole moments in Boltzmann hierarchy
    num_G_ell_modes : int
        Number of polarization multipole moments in Boltzmann hierarchy

    Methods:
    --------
    rho : Compute photon density (units: eV cm^{-3})
    P : Compute photon pressure (units: eV cm^{-3})
    y_ini : Compute initial perturbation conditions 
    y_prime : Compute perturbation time derivatives 
    """
    num_F_ell_modes : int = eqx.field(static=True)
    num_G_ell_modes : int = eqx.field(static=True)
    name = "Photon"

    def __init__(self, delta_idx, specs):
        super().__init__(delta_idx, specs)
        self.name = "Photon"
        self.num_F_ell_modes = specs["l_max_g"] + 1
        self.num_G_ell_modes = specs["l_max_pol_g"] + 1
        self.num_moments = self.num_F_ell_modes + self.num_G_ell_modes

    def rho(self, lna, args):
        """
        Compute photon density.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Photon density (units: eV cm^{-3})
        """
        params = args
        a = jnp.exp(lna)
        return jnp.pi**2/15. * params["TCMB0"]**4 / a**4 / (cnst.c * cnst.hbar)**3

    def P(self, lna, args):
        """
        Compute photon pressure.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        float
            Photon pressure (units: eV cm^{-3})
        """
        params = args
        return self.rho(lna, params)/3.

    def y_ini(self, k, tau_ini, args):
        """
        Compute initial conditions for photon perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        tau_ini : float
            Initial conformal time (units: Mpc)
        args : dict
            Cosmological parameters (params)

        Returns:
        --------
        array
            Initial perturbation mode values (units: 1/Mpc for theta, else dimensionless)
        """
        params = args
        delta = - (k*tau_ini)**2/3. * (1.-params["om"]*tau_ini/5.)
        theta = - k**4 * tau_ini**3/36. * (1.-3.*(1.+5.*params['R_b']-params['R_nu'])/20./(1.-params['R_nu'])*params["om"]*tau_ini)
        return jnp.concatenate((jnp.array([delta, theta]), jnp.zeros(self.num_moments - 2)))

    def y_prime(self, k, lna, metric_h_prime, metric_eta_prime, y, args):
        """
        Compute time derivatives of photon perturbations.

        Parameters:
        -----------
        k : float
            Wavenumber (units: Mpc^{-1})
        lna : float
            Logarithm of scale factor
        metric_h_prime : float
            Derivative of metric h
        metric_eta_prime : float
            Derivative of metric eta
        y : array
            Current perturbation mode values
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        array
            Time derivatives of perturbation modes (units: 1/Mpc for theta, else dimensionless)
        """
        BG, params, species_list, species_dict = args
        # Get Baryon from list
        i = species_dict["Baryon"]
        baryon = species_list[i]

        aH    = BG.aH(lna, params)
        tau_c = BG.tau_c(lna, params)
        tau   = BG.tau(lna)

        Flmax = self.num_F_ell_modes-1
        Glmax = self.num_G_ell_modes-1
        F = lax.dynamic_slice(y, (self.delta_idx,), (self.num_F_ell_modes,))
        G = lax.dynamic_slice(y, (self.delta_idx+self.num_F_ell_modes,), (self.num_G_ell_modes,))
        delta = F[0]
        theta = F[1]
        sigma = F[2]
        theta_b = y[baryon.delta_idx+1]

        delta_prime = -4./3./aH*theta - 2./3.*metric_h_prime
        theta_prime = k**2/aH*(delta/4.-sigma) + (theta_b-theta)/aH/tau_c
        sigma_prime = 4./15./aH*theta - 3./10.*k/aH*F[3] + 2./15.*metric_h_prime + 4./5.*metric_eta_prime - 9./10./aH/tau_c*sigma + (G[0]+G[2])/20./aH/tau_c
        F3_prime    = k/7./aH * (6.*sigma - 4.*F[4]) - F[3]/aH/tau_c

        # Temperature Boltzmann Hierarchy
        L = jnp.arange(4, Flmax) # Excludes the lmax mode
        Fl_prime    = 1./(2.*L+1.)*k/aH * (L*F[L-1]-(L+1)*F[L+1]) - F[L]/aH/tau_c
        Flmax_prime = k/aH*F[Flmax-1] - (Flmax+1)/aH/tau*F[Flmax] - F[Flmax]/aH/tau_c

        # Polarization Boltzmann Hierarchy
        L = jnp.arange(0, Glmax) # Excludes the lmax mode
        Gl_prime    = 1./(2.*L+1.)*k/aH * (L*G[L-1]-(L+1)*G[L+1]) - G[L]/aH/tau_c \
                    + (2.*sigma+G[0]+G[2])/2./aH/tau_c * jnp.concatenate((jnp.array([1., 0., 0.2]), jnp.zeros(Glmax-3)))

        Glmax_prime = k/aH*G[Glmax-1] - (Glmax+1)/aH/tau*G[Glmax] - G[Glmax]/aH/tau_c
        return jnp.concatenate((jnp.array([delta_prime, theta_prime, sigma_prime, F3_prime]), Fl_prime, jnp.array([Flmax_prime]), Gl_prime, jnp.array([Glmax_prime])))